#!/usr/bin/python

# Copyright (c) 2013 GRNET S.A., SRCE, IN2P3 CNRS Computing Centre
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of either GRNET S.A., SRCE or IN2P3 CNRS Computing
# Centre
#
# The work represented by this source file is partially funded by
# the EGI-InSPIRE project through the European Commission's 7th
# Framework Programme (contract # INFSO-RI-261323)

import argparse
import copy
import datetime
import os
import sys
import xml.dom.minidom

from argo_egi_connectors.config import Global, CustomerConf
from argo_egi_connectors.helpers import gen_fname_repdate, make_connection, module_class_name, parse_xml, ConnectorError, write_state
from argo_egi_connectors.writers import AvroWriter
from argo_egi_connectors.writers import SingletonLogger as Logger
from urlparse import urlparse

logger = None

SERVENDPI = '/gocdbpi/private/?method=get_service_endpoint'
SITESPI = '/gocdbpi/private/?method=get_site'
SERVGROUPPI = '/gocdbpi/private/?method=get_service_group'

LegMapServType = {'SRM' : 'SRMv2'}
fetchtype = ''

globopts = {}
custname = ''

isok = True

class GOCDBReader:
    def __init__(self, feed, scopes):
        self._o = urlparse(feed)
        self.scopes = scopes if scopes else set(['NoScope'])
        for scope in self.scopes:
            code = "self.serviceList%s = dict(); " % scope
            code += "self.groupList%s = dict();" % scope
            code += "self.siteList%s = dict()" % scope
            exec code
        self.fetched = False
        self.state = True

    def getGroupOfServices(self):
        if not self.fetched:
            if not self.state or not self.loadDataIfNeeded():
                return []

        groups, gl = list(), list()

        for scope in self.scopes:
            code = "gl = gl + [value for key, value in self.groupList%s.iteritems()]" % scope
            exec code

        for d in gl:
            for service in d['services']:
                g = dict()
                g['type'] = fetchtype.upper()
                g['group'] = d['name']
                g['service'] = service['type']
                g['hostname'] = service['hostname']
                g['group_monitored'] = d['monitored']
                g['tags'] = {'scope' : d['scope'], \
                            'monitored' : '1' if service['monitored'] == "Y" else '0', \
                            'production' : '1' if service['production'] == "Y" else '0'}
                groups.append(g)

        return groups

    def getGroupOfGroups(self):
        if not self.fetched:
            if not self.state or not self.loadDataIfNeeded():
                return []

        groupofgroups, gl = list(), list()

        if fetchtype == "ServiceGroups":
            for scope in self.scopes:
                code = "gl = gl + [value for key, value in self.groupList%s.iteritems()]" % scope
                exec code
            for d in gl:
                g = dict()
                g['type'] = 'PROJECT'
                g['group'] = custname
                g['subgroup'] = d['name']
                g['tags'] = {'monitored' : '1' if d['monitored'] == 'Y' else '0',
                            'scope' : d['scope']}
                groupofgroups.append(g)
        else:
            gg = []
            for scope in self.scopes:
                code = "gg = gg + sorted([value for key, value in self.siteList%s.iteritems()], key=lambda s: s['ngi'])" % scope
                exec code

            for gr in gg:
                g = dict()
                g['type'] = 'NGI'
                g['group'] = gr['ngi']
                g['subgroup'] = gr['site']
                g['tags'] = {'certification' : gr['certification'], \
                             'scope' : gr['scope'], \
                             'infrastructure' : gr['infrastructure']}

                groupofgroups.append(g)

        return groupofgroups

    def getGroupOfEndpoints(self):
        if not self.fetched:
            if not self.state or not self.loadDataIfNeeded():
                return []

        groupofendpoints, ge = list(), list()
        for scope in self.scopes:
            code = "ge = ge + sorted([value for key, value in self.serviceList%s.iteritems()], key=lambda s: s['site'])" % scope
            exec code

        for gr in ge:
            g = dict()
            g['type'] = fetchtype.upper()
            g['group'] = gr['site']
            g['service'] = gr['type']
            g['hostname'] = gr['hostname']
            g['tags'] = {'scope' : gr['scope'], \
                         'monitored' : '1' if gr['monitored'] == "Y" else '0', \
                         'production' : '1' if gr['production'] == "Y" else '0'}
            groupofendpoints.append(g)

        return groupofendpoints

    def loadDataIfNeeded(self):
        scopequery = "'&scope='+scope"
        for scope in self.scopes:
            try:
                eval("self.getSitesInternal(self.siteList%s, %s)" % (scope, '' if scope == 'NoScope' else scopequery))
                eval("self.getServiceGroups(self.groupList%s, %s)" % (scope, '' if scope == 'NoScope' else scopequery))
                eval("self.getServiceEndpoints(self.serviceList%s, %s)" % (scope, '' if scope == 'NoScope' else scopequery))
                self.fetched = True
            except Exception:
                self.state = False
                return False

        return True

    def _get_xmldata(self, scope, pi):
        res = make_connection(logger, globopts, self._o.scheme, self._o.netloc,
                                pi + scope,
                                module_class_name(self))
        doc = parse_xml(logger, res, self._o.scheme + '://' + self._o.netloc + pi,
                        module_class_name(self))
        return doc

    def getServiceEndpoints(self, serviceList, scope):
        try:
            doc = self._get_xmldata(scope, SERVENDPI)
            services = doc.getElementsByTagName('SERVICE_ENDPOINT')
            for service in services:
                serviceId = ''
                if service.getAttributeNode('PRIMARY_KEY'):
                    serviceId = str(service.attributes['PRIMARY_KEY'].value)
                if serviceId not in serviceList:
                    serviceList[serviceId] = {}
                serviceList[serviceId]['hostname'] = service.getElementsByTagName('HOSTNAME')[0].childNodes[0].data
                serviceList[serviceId]['type'] = service.getElementsByTagName('SERVICE_TYPE')[0].childNodes[0].data
                serviceList[serviceId]['monitored'] = service.getElementsByTagName('NODE_MONITORED')[0].childNodes[0].data
                serviceList[serviceId]['production'] = service.getElementsByTagName('IN_PRODUCTION')[0].childNodes[0].data
                serviceList[serviceId]['site'] = service.getElementsByTagName('SITENAME')[0].childNodes[0].data
                serviceList[serviceId]['roc'] = service.getElementsByTagName('ROC_NAME')[0].childNodes[0].data
                serviceList[serviceId]['scope'] = scope.split('=')[1]
                serviceList[serviceId]['sortId'] = serviceList[serviceId]['hostname'] + '-' + serviceList[serviceId]['type'] + '-' + serviceList[serviceId]['site']

        except ConnectorError as e:
            raise e

        except (KeyError, IndexError, TypeError, AttributeError, AssertionError) as e:
            logger.error(module_class_name(self) + ': Error parsing feed %s - %s' % (self._o.scheme + '://' + self._o.netloc + SERVENDPI,
                                                                                     repr(e).replace('\'','').replace('\"', '')))
            raise e

    def getSitesInternal(self, siteList, scope):
        try:
            doc = self._get_xmldata(scope, SITESPI)
            sites = doc.getElementsByTagName('SITE')
            for site in sites:
                siteName = site.getAttribute('NAME')
                if siteName not in siteList:
                    siteList[siteName] = {'site': siteName}
                siteList[siteName]['infrastructure'] = site.getElementsByTagName('PRODUCTION_INFRASTRUCTURE')[0].childNodes[0].data
                siteList[siteName]['certification'] = site.getElementsByTagName('CERTIFICATION_STATUS')[0].childNodes[0].data
                siteList[siteName]['ngi'] = site.getElementsByTagName('ROC')[0].childNodes[0].data
                siteList[siteName]['scope'] = scope.split('=')[1]

        except ConnectorError as e:
            raise e

        except (KeyError, IndexError, TypeError, AttributeError, AssertionError) as e:
            logger.error(module_class_name(self) + ': Error parsing feed %s - %s' % (self._o.scheme + '://' + self._o.netloc + SITESPI,
                                                                                     repr(e).replace('\'','').replace('\"', '')))
            raise e

    def getServiceGroups(self, groupList, scope):
        try:
            doc = self._get_xmldata(scope, SERVGROUPPI)
            groups = doc.getElementsByTagName('SERVICE_GROUP')
            for group in groups:
                groupId = group.getAttribute('PRIMARY_KEY')
                if groupId not in groupList:
                    groupList[groupId] = {}
                groupList[groupId]['name'] = group.getElementsByTagName('NAME')[0].childNodes[0].data
                groupList[groupId]['monitored'] = group.getElementsByTagName('MONITORED')[0].childNodes[0].data
                groupList[groupId]['scope'] = scope.split('=')[1]
                groupList[groupId]['services'] = []
                services = group.getElementsByTagName('SERVICE_ENDPOINT')
                for service in services:
                    serviceDict = {}
                    serviceDict['hostname'] = service.getElementsByTagName('HOSTNAME')[0].childNodes[0].data
                    serviceDict['type'] = service.getElementsByTagName('SERVICE_TYPE')[0].childNodes[0].data
                    serviceDict['monitored'] = service.getElementsByTagName('NODE_MONITORED')[0].childNodes[0].data
                    serviceDict['production'] = service.getElementsByTagName('IN_PRODUCTION')[0].childNodes[0].data
                    groupList[groupId]['services'].append(serviceDict)

        except ConnectorError as e:
            raise e

        except (KeyError, IndexError, TypeError, AttributeError, AssertionError) as e:
            logger.error(module_class_name(self) + ': Error parsing feed %s - %s' % (self._o.scheme + '://' + self._o.netloc + SERVGROUPPI,
                                                                                     repr(e).replace('\'','').replace('\"', '')))
            raise e

def extract_siteorngi_tags(tag, ggtags):
    gg, key, exval = None, None, None
    if tag.lower() in [t.lower() for t in ggtags.iterkeys()]:
        for k, v in ggtags.iteritems():
            if tag.lower() in k.lower():
                gg = ggtags[k]
                key = k
        exval = ggtags.pop(key)
        if isinstance(gg, list):
            gg = [t.lower() for t in gg]
        else:
            gg = gg.lower()

    if exval:
        return gg, dict({key: exval})
    else:
        return gg, None

def filter_by_tags(tags, listofelem, subgroups=None):
    for attr in tags.keys():
        def getit(elem):
            value = elem['tags'][attr.lower()]
            if value == '1': value = 'Y'
            elif value == '0': value = 'N'
            if isinstance(tags[attr], list):
                for a in tags[attr]:
                    if subgroups and elem['group'] in subgroups and \
                            value.lower() == a.lower():
                        return True
                    elif not subgroups and value.lower() == a.lower():
                        return True
            else:
                if subgroups and elem['group'] in subgroups and \
                        value.lower() == tags[attr].lower():
                    return True
                elif not subgroups and value.lower() == tags[attr].lower():
                    return True

        try:
            listofelem = filter(getit, listofelem)
        except KeyError as e:
            logger.error('Wrong tags specified: %s' % e)
    return listofelem

def main():
    global logger, globopts
    parser = argparse.ArgumentParser(description="""Fetch entities (ServiceGroups, Sites, Endpoints)
                                                    from GOCDB for every customer and job listed in customer.conf and write them
                                                    in an appropriate place""")
    parser.add_argument('-c', dest='custconf', nargs=1, metavar='customer.conf', help='path to customer configuration file', type=str, required=False)
    parser.add_argument('-g', dest='gloconf', nargs=1, metavar='global.conf', help='path to global configuration file', type=str, required=False)
    args = parser.parse_args()
    group_endpoints, group_groups = [], []

    logger = Logger(os.path.basename(sys.argv[0]))

    certs = {'Authentication': ['HostKey', 'HostCert', 'CAPath', 'CAFile', 'VerifyServerCert']}
    schemas = {'AvroSchemas': ['TopologyGroupOfEndpoints', 'TopologyGroupOfGroups']}
    output = {'Output': ['TopologyGroupOfEndpoints', 'TopologyGroupOfGroups']}
    conn = {'Connection': ['Timeout', 'Retry']}
    state = {'InputState': ['SaveDir', 'Days']}
    confpath = args.gloconf[0] if args.gloconf else None
    cglob = Global(confpath, certs, schemas, output, conn, state)
    globopts = cglob.parse()

    confpath = args.custconf[0] if args.custconf else None
    confcust = CustomerConf(sys.argv[0], confpath)
    confcust.parse()
    confcust.make_dirstruct()
    confcust.make_dirstruct(globopts['InputStateSaveDir'.lower()])
    feeds = confcust.get_mapfeedjobs(sys.argv[0], 'GOCDB', deffeed='https://goc.egi.eu/gocdbpi/')

    for feed, jobcust in feeds.items():
        scopes = confcust.get_feedscopes(feed, jobcust)
        gocdb = GOCDBReader(feed, scopes)

        for job, cust in jobcust:
            jobdir = confcust.get_fulldir(cust, job)
            jobstatedir = confcust.get_fullstatedir(globopts['InputStateSaveDir'.lower()], cust, job)

            global fetchtype, custname
            fetchtype = confcust.get_gocdb_fetchtype(job)
            custname = confcust.get_custname(cust)

            if fetchtype == 'ServiceGroups':
                group_endpoints = gocdb.getGroupOfServices()
            else:
                group_endpoints = gocdb.getGroupOfEndpoints()
            group_groups = gocdb.getGroupOfGroups()

            write_state(sys.argv[0], jobstatedir, gocdb.state, globopts['InputStateDays'.lower()])

            if not gocdb.state:
                continue

            numge = len(group_endpoints)
            numgg = len(group_groups)

            ggtags = confcust.get_gocdb_ggtags(job)
            ggngi, ggsite = None, None
            dngi, dsite = None, None

            ggngi, dngi = extract_siteorngi_tags('ngi', ggtags)
            ggsite, dsite = extract_siteorngi_tags('site', ggtags)

            if ggngi:
                group_groups = filter(lambda e: e['group'].lower() in ggngi, group_groups)

            if ggsite:
                group_groups = filter(lambda e: e['subgroup'].lower() in ggsite, group_groups)

            if ggtags:
                group_groups = filter_by_tags(ggtags, group_groups)

            if dngi:
                ggtags.update(dngi)

            if dsite:
                ggtags.update(dsite)

            allsubgroups = set([e['subgroup'] for e in group_groups])

            filename = gen_fname_repdate(logger, globopts['OutputTopologyGroupOfGroups'.lower()], jobdir)
            avro = AvroWriter(globopts['AvroSchemasTopologyGroupOfGroups'.lower()], filename,
                            group_groups, os.path.basename(sys.argv[0]))
            avro.write()

            gelegmap = []
            for g in group_endpoints:
                if g['service'] in LegMapServType.keys():
                    gelegmap.append(copy.copy(g))
                    gelegmap[-1]['service'] = LegMapServType[g['service']]
            getags = confcust.get_gocdb_getags(job)
            numgeleg = len(gelegmap)
            if getags or allsubgroups:
                group_endpoints = filter_by_tags(getags, group_endpoints, allsubgroups)
                gelegmap = filter_by_tags(getags, gelegmap, allsubgroups)

            filename = gen_fname_repdate(logger, globopts['OutputTopologyGroupOfEndpoints'.lower()], jobdir)
            avro = AvroWriter(globopts['AvroSchemasTopologyGroupOfEndpoints'.lower()], filename,
                            group_endpoints + gelegmap, os.path.basename(sys.argv[0]))
            avro.write()

            logger.info('Customer:'+custname+' Job:'+job+' Fetched Endpoints:%d' % (numge + numgeleg) +' Groups(%s):%d' % (fetchtype, numgg))
            if getags or ggtags:
                selstr = 'Customer:%s Job:%s Selected ' % (custname, job)
                selge, selgg = '', ''
                if getags:
                    for key, value in getags.items():
                        if isinstance(value, list):
                            value = '['+','.join(value)+']'
                        selge += '%s:%s,' % (key, value)
                    selstr += 'Endpoints(%s):' % selge[:len(selge) - 1]
                    selstr += '%d ' % (len(group_endpoints) + len(gelegmap))
                if ggtags:
                    for key, value in ggtags.items():
                        if isinstance(value, list):
                            value = '['+','.join(value)+']'
                        selgg += '%s:%s,' % (key, value)
                    selstr += 'Groups(%s):' % selgg[:len(selgg) - 1]
                    selstr += '%d' % (len(group_groups))

                logger.info(selstr)
main()
